/*
 * This file is part of Cerealization.
 *
 * Copyright (c) 2013 Spout LLC <http://www.spout.org/>
 * Cerealization is licensed under the Spout License Version 1.
 *
 * Cerealization is free software: you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation, either version 3 of the License, or (at your option)
 * any later version.
 *
 * In addition, 180 days after any changes are published, you can use the
 * software, incorporating those changes, under the terms of the MIT license,
 * as described in the Spout License Version 1.
 *
 * Cerealization is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for
 * more details.
 *
 * You should have received a copy of the GNU Lesser General Public License,
 * the MIT license and the Spout License Version 1 along with this program.
 * If not, see <http://www.gnu.org/licenses/> for the GNU Lesser General Public
 * License and see <http://spout.in/licensev1> for the full license, including
 * the MIT license.
 */
package org.spout.cereal.config.serialization;

import java.lang.reflect.GenericArrayType;
import java.lang.reflect.ParameterizedType;
import java.lang.reflect.Type;
import java.lang.reflect.WildcardType;

/**
 * A wrapper around {@link Type} to allow easier access to parameterized types.
 */
public class GenericType {
	private final Class<?> mainType;
	private final GenericType[] neededGenerics;
	private final GenericType arrayType;
	private final Type rawType;

	public GenericType(Type rawType) {
		if (rawType instanceof ParameterizedType) {
			ParameterizedType type = (ParameterizedType) rawType;
			mainType = getMainType(type.getRawType());
			neededGenerics = toGenericType(type.getActualTypeArguments());
			arrayType = null;
		} else if (rawType instanceof GenericArrayType) {
			GenericArrayType type = (GenericArrayType) rawType;
			mainType = null;
			neededGenerics = new GenericType[0];
			arrayType = new GenericType(type.getGenericComponentType());
		} else {
			mainType = getMainType(rawType);
			neededGenerics = new GenericType[0];
			arrayType = null;
		}
		this.rawType = rawType;
	}

	public Class<?> getMainType() {
		return mainType;
	}

	public GenericType[] getGenerics() {
		return neededGenerics;
	}

	public Type getRawType() {
		return rawType;
	}

	public boolean isArray() {
		return arrayType != null;
	}

	public GenericType getArrayType() {
		return arrayType;
	}

	private static Class<?> getMainType(Type type) {
		if (type instanceof Class) {
			return (Class<?>) type;
		} else if (type instanceof WildcardType) {
			return (Class<?>) ((WildcardType) type).getUpperBounds()[0];
		} else {
			throw new IllegalArgumentException("Unknown main class type: " + type.getClass());
		}
	}

	public static GenericType[] toGenericType(Type[] types) {
		GenericType[] genericTypes = new GenericType[types.length];
		for (int i = 0; i < types.length; ++i) {
			genericTypes[i] = new GenericType(types[i]);
		}
		return genericTypes;
	}
}
